@@ -30,6 +30,7 @@
  */
 
 const gulp = require("gulp");
+const gutil = require("gulp-util");
 const map = require("map-stream");
 const zip = require("gulp-zip");
 const gzip = require("gulp-gzip");
@@ -40,76 +41,81 @@ const concat = require("gulp-concat");
 const ts = require("gulp-typescript");
 const debug = require("gulp-debug");
 const glob = require("glob");
-const jsonTransform = require('gulp-json-transform');
+const jsonTransform = require("gulp-json-transform");
 const fs = require("fs");
-const del = require("del");
-const through = require('through2');
-const File = require('vinyl');
-const Stream = require('stream').Stream;
-const vfs = require('vinyl-fs');
+const through = require("through2");
+const File = require("vinyl");
+const Stream = require("stream").Stream;
+const vfs = require("vinyl-fs");
+const webpack = require("webpack");
+const po2json = require("po2json");
+const path = require("path");
 
 const paths = {
   ts: {
     src: [
-      "src/**/*.{ts,tsx}",
+      "src/**/*.{ts,tsx,js}",
       "!src/**/*-test*.ts",
     ],
     decl: [
-      "decl/lib.es6.d.ts",
-      "decl/urijs/URIjs.d.ts",
-      "decl/systemjs/systemjs.d.ts",
-      "decl/react-global.d.ts",
+      "decl/jed.d.ts",
       "decl/chrome/chrome.d.ts",
+      "decl/urijs.d.ts",
     ],
     test: [
-        "testlib/**/.ts",
         "src/**/*-test*.ts",
     ],
   },
   // distributed in the chrome extension
   dist: [
+    "dist/*-bundle.js",
+    "dist/*-bundle.js.map",
+    "emscripten/taler-emscripten-lib.js",
     "img/icon.png",
     "img/logo.png",
-    "src/**/*.{js,css,html}",
-    "testlib/**/*.{js,ts,tsx,html}",
+    "src/**/*.{css,html}",
   ],
   // for the source distribution
   extra: [
-      "scripts/prove-node",
-      "scripts/prove-selenium",
-      "src/i18n/*.po",
-      "src/i18n/*.pot",
-      "decl/**/*.d.ts",
       "AUTHORS",
-      "README",
       "COPYING",
       "Makefile",
+      "README",
       "configure",
+      "decl/**/*.d.ts",
       "gulpfile.js",
-      "tsconfig.json",
+      "manifest.json",
       "package.json",
+      "pogen/example/test.ts",
       "pogen/pogen.ts",
       "pogen/tsconfig.json",
-      "pogen/example/test.ts",
-      // Only in extra, because the manifest is processed/generated
-      // targets other than "srcdist".
-      "manifest.json",
+      "src/i18n/*.po",
+      "src/i18n/*.pot",
+      "src/i18n/poheader",
+      "src/i18n/strings-prelude",
+      "tooling/**",
+      "tsconfig.json",
+      "webpack.config.js",
   ],
 };
 
 
 const tsBaseArgs = {
-  target: "es6",
+  target: "es5",
   jsx: "react",
   reactNamespace: "React",
   experimentalDecorators: true,
-  module: "system",
+  module: "commonjs",
   sourceMap: true,
-  noLib: true,
+  lib: ["ES6", "DOM"],
   noImplicitReturns: true,
   noFallthroughCasesInSwitch: true,
-  strictNullChecks: true,
+  strict: true,
+  outDir: "build/src/",
   noImplicitAny: true,
+  allowJs: true,
+  checkJs: true,
+  noUnusedLocals: true,
 };
 
 
@@ -151,35 +157,23 @@ function concatStreams (/*streams...*/) {
 }
 
 
-gulp.task("clean", function () {
-  return del("build/ext");
-});
 
-
-gulp.task("dist-prod", ["clean"], function () {
+gulp.task("dist-prod", ["compile-prod"], function () {
   return vfs.src(paths.dist, {base: ".", stripBOM: false})
              .pipe(gulp.dest("build/ext/"));
 });
 
-gulp.task("compile-prod", ["clean"], function () {
-  const tsArgs = {};
-  Object.assign(tsArgs, tsBaseArgs);
-  tsArgs.typescript = require("typescript");
-  // relative to the gulp.dest
-  tsArgs.outDir = ".";
-  // We don't want source maps for production
-  tsArgs.sourceMap = undefined;
-  let opts = {base: "."};
-  const files = concatStreams(
-          gulp.src(paths.ts.src, opts),
-          gulp.src(paths.ts.decl, opts));
-
-  return files
-      .pipe(ts(tsArgs))
-      .pipe(gulp.dest("build/ext/"));
+gulp.task("compile-prod", function (callback) {
+  let config = require("./webpack.config.js")({prod: true});
+  webpack(config, function(err, stats) {
+    if(err) {
+      throw new gutil.PluginError("webpack", err);
+    }
+    callback();
+  });
 });
 
-gulp.task("manifest-stable", ["clean"], function () {
+gulp.task("manifest-stable", function () {
   return gulp.src("manifest.json")
              .pipe(jsonTransform((data) => {
                data.name = "GNU Taler Wallet";
@@ -188,7 +182,7 @@ gulp.task("manifest-stable", ["clean"], function () {
              .pipe(gulp.dest("build/ext/"));
 });
 
-gulp.task("manifest-unstable", ["clean"], function () {
+gulp.task("manifest-unstable", function () {
   return gulp.src("manifest.json")
              .pipe(jsonTransform((data) => {
                data.name = "GNU Taler Wallet (unstable)";
@@ -198,7 +192,7 @@ gulp.task("manifest-unstable", ["clean"], function () {
 });
 
 
-gulp.task("package-stable", ["compile-prod", "dist-prod", "manifest-stable"], function () {
+gulp.task("package-stable", ["dist-prod", "manifest-stable"], function () {
   let basename = String.prototype.concat("taler-wallet-stable-", manifest.version_name, "-", manifest.version);
   let zipname = basename + ".zip";
   let xpiname = basename + ".xpi";
@@ -208,7 +202,7 @@ gulp.task("package-stable", ["compile-prod", "dist-prod", "manifest-stable"], fu
              .pipe(symlink("build/" + xpiname, {relative: true, force: true}));
 });
 
-gulp.task("package-unstable", ["compile-prod", "dist-prod", "manifest-unstable"], function () {
+gulp.task("package-unstable", ["dist-prod", "manifest-unstable"], function () {
   let basename = String.prototype.concat("taler-wallet-unstable-", manifest.version_name, "-",  manifest.version);
   let zipname = basename + ".zip";
   let xpiname = basename + ".xpi";
@@ -255,7 +249,7 @@ gulp.task("pogenjs", [], function () {
 /**
  * Extract .po files from source code
  */
-gulp.task("pogen", ["pogenjs"], function (cb) {
+gulp.task("pogen", function (cb) {
   throw Error("not yet implemented");
 });
 
@@ -271,10 +265,10 @@ function tsconfig(confBase) {
     files: []
   };
   Object.assign(conf.compilerOptions, confBase);
-  return through.obj(function(file, enc, cb) {
+  return through.obj(function (file, enc, cb) {
     conf.files.push(file.relative);
     cb();
-  }, function(cb) {
+  }, function (cb) {
     conf.files.sort();
     let x = JSON.stringify(conf, null, 2);
     let f = new File({
@@ -287,9 +281,69 @@ function tsconfig(confBase) {
 }
 
 
+/**
+ * Get the content of a Vinyl file object as a buffer.
+ */
+function readContentsBuffer(file, cb) {
+  if (file.isBuffer()) {
+    cb(file.contents);
+    return;
+  }
+  if (!file.isStream()) {
+    throw Error("file must be stream or buffer");
+  }
+  const chunks = [];
+  file.contents.on("data", function (chunk) {
+    if (!Buffer.isBuffer(chunk)) {
+      throw Error("stream data must be a buffer");
+    }
+    chunks.push(chunk);
+  });
+  file.contents.on("end", function (chunk) {
+    cb(Buffer.concat(chunks));
+  });
+  file.contents.on("error", function (err) {
+    cb(undefined, err);
+  });
+}
+
+
+/**
+ * Combine multiple translations (*.po files) into
+ * one JavaScript file.
+ */
+function pofilesToJs(targetPath) {
+  const outStream = through();
+  const f = new File({
+    path: targetPath,
+    contents: outStream,
+  });
+  const prelude = fs.readFileSync("./src/i18n/strings-prelude");
+  outStream.write(prelude);
+  return through.obj(function (file, enc, cb) {
+    readContentsBuffer(file, function (buf, error) {
+      if (error) {
+        throw error;
+      }
+      const lang = path.basename(file.path, ".po");
+      if (!lang) {
+        throw Error();
+      }
+      console.log("lang", lang);
+      const pojson = po2json.parse(buf, {format: "jed1.x", fuzzy: true});
+      outStream.write("strings['" + lang + "'] = " + JSON.stringify(pojson, null, "  ") + ";\n");
+      cb();
+    });
+  }, function (cb) {
+    this.push(f);
+    return cb();
+  });
+}
+
+
 // Generate the tsconfig file
 // that should be used during development.
-gulp.task("tsconfig", function() {
+gulp.task("tsconfig", function () {
   let opts = {base: "."};
   const files = concatStreams(
           vfs.src(paths.ts.src, opts),
@@ -299,5 +353,11 @@ gulp.task("tsconfig", function() {
               .pipe(gulp.dest("."));
 });
 
+gulp.task("po2js", function () {
+  return gulp.src("src/i18n/*.po", {base: "."})
+             .pipe(pofilesToJs("src/i18n/strings.ts"))
+             .pipe(gulp.dest("."));
+});
+
 
 gulp.task("default", ["package-stable", "tsconfig"]);
