@@ -21,43 +21,54 @@
  * @author Florian Dold
  */
 
-"use strict";
 
-
-export interface JoinResult<L,R> {
+/**
+ * Result of an inner join.
+ */
+export interface JoinResult<L, R> {
   left: L;
   right: R;
 }
 
-export interface JoinLeftResult<L,R> {
+/**
+ * Result of a left outer join.
+ */
+export interface JoinLeftResult<L, R> {
   left: L;
   right?: R;
 }
 
 
+/**
+ * Definition of an object store.
+ */
 export class Store<T> {
-  name: string;
-  validator?: (v: T) => T;
-  storeParams?: IDBObjectStoreParameters;
-
-  constructor(name: string, storeParams?: IDBObjectStoreParameters,
-              validator?: (v: T) => T) {
-    this.name = name;
-    this.validator = validator;
-    this.storeParams = storeParams;
+  constructor(public name: string,
+              public storeParams?: IDBObjectStoreParameters,
+              public validator?: (v: T) => T) {
   }
 }
 
-export class Index<S extends IDBValidKey,T> {
-  indexName: string;
-  storeName: string;
-  keyPath: string | string[];
 
-  constructor(s: Store<T>, indexName: string, keyPath: string | string[]) {
+/**
+ * Definition of an index.
+ */
+export class Index<S extends IDBValidKey, T> {
+  /**
+   * Name of the store that this index is associated with.
+   */
+  storeName: string;
+
+  constructor(s: Store<T>, public indexName: string, public keyPath: string | string[]) {
     this.storeName = s.name;
-    this.indexName = indexName;
-    this.keyPath = keyPath;
   }
+
+  /**
+   * We want to have the key type parameter in use somewhere,
+   * because otherwise the compiler complains.  In iterIndex the
+   * key type is pretty useful.
+   */
+  protected _dummyKey: S|undefined;
 }
 
 /**
@@ -65,28 +76,155 @@ export class Index<S extends IDBValidKey,T> {
  * with indices.
  */
 export interface QueryStream<T> {
-  indexJoin<S,I extends IDBValidKey>(index: Index<I,S>,
-                                     keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>>;
-  indexJoinLeft<S,I extends IDBValidKey>(index: Index<I,S>,
-                                     keyFn: (obj: T) => I): QueryStream<JoinLeftResult<T, S>>;
-  keyJoin<S,I extends IDBValidKey>(store: Store<S>,
-                                   keyFn: (obj: T) => I): QueryStream<JoinResult<T,S>>;
-  filter(f: (T: any) => boolean): QueryStream<T>;
+  /**
+   * Join the current query with values from an index.
+   * The left side of the join is extracted via a function from the stream's
+   * result, the right side of the join is the key of the index.
+   */
+  indexJoin<S, I extends IDBValidKey>(index: Index<I, S>, keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>>;
+  /**
+   * Join the current query with values from an index, and keep values in the
+   * current stream that don't have a match.  The left side of the join is
+   * extracted via a function from the stream's result, the right side of the
+   * join is the key of the index.
+   */
+  indexJoinLeft<S, I extends IDBValidKey>(index: Index<I, S>,
+                                          keyFn: (obj: T) => I): QueryStream<JoinLeftResult<T, S>>;
+  /**
+   * Join the current query with values from another object store.
+   * The left side of the join is extracted via a function over the current query,
+   * the right side of the join is the key of the object store.
+   */
+  keyJoin<S, I extends IDBValidKey>(store: Store<S>, keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>>;
+
+  /**
+   * Only keep elements in the result stream for which the predicate returns
+   * true.
+   */
+  filter(f: (x: T) => boolean): QueryStream<T>;
+
+  /**
+   * Reduce the stream, resulting in a single value.
+   */
   reduce<S>(f: (v: T, acc: S) => S, start?: S): Promise<S>;
-  map<S>(f: (x:T) => S): QueryStream<S>;
+
+  /**
+   * Map each element of the stream using a function, resulting in another
+   * stream of a different type.
+   */
+  map<S>(f: (x: T) => S): QueryStream<S>;
+
+  /**
+   * Map each element of the stream to a potentially empty array, and collect
+   * the result in a stream of the flattened arrays.
+   */
   flatMap<S>(f: (x: T) => S[]): QueryStream<S>;
+
+  /**
+   * Collect the stream into an array and return a promise for it.
+   */
   toArray(): Promise<T[]>;
 
+  /**
+   * Get the first value of the stream.
+   */
+  first(): QueryValue<T>;
+
   then(onfulfill: any, onreject: any): any;
 }
 
-export let AbortTransaction = Symbol("abort_transaction");
+
+/**
+ * Query result that consists of at most one value.
+ */
+export interface QueryValue<T> {
+  /**
+   * Apply a function to a query value.
+   */
+  map<S>(f: (x: T) => S): QueryValue<S>;
+  /**
+   * Conditionally execute either of two queries based
+   * on a property of this query value.
+   *
+   * Useful to properly implement complex queries within a transaction (as
+   * opposed to just computing the conditional and then executing either
+   * branch).  This is necessary since IndexedDB does not allow long-lived
+   * transactions.
+   */
+  cond<R>(f: (x: T) => boolean, onTrue: (r: QueryRoot) => R, onFalse: (r: QueryRoot) => R): Promise<void>;
+}
+
+
+abstract class BaseQueryValue<T> implements QueryValue<T> {
+
+  constructor(public root: QueryRoot) {
+  }
+
+  map<S>(f: (x: T) => S): QueryValue<S> {
+    return new MapQueryValue<T, S>(this, f);
+  }
+
+  cond<R>(f: (x: T) => boolean, onTrue: (r: QueryRoot) => R, onFalse: (r: QueryRoot) => R): Promise<void> {
+    return new Promise<void>((resolve, reject) => {
+      this.subscribeOne((v, tx) => {
+        if (f(v)) {
+          onTrue(new QueryRoot(this.root.db));
+        } else {
+          onFalse(new QueryRoot(this.root.db));
+        }
+      });
+      resolve();
+    });
+  }
+
+  abstract subscribeOne(f: SubscribeOneFn): void;
+}
+
+class FirstQueryValue<T> extends BaseQueryValue<T> {
+  private gotValue = false;
+  private s: QueryStreamBase<T>;
+
+  constructor(stream: QueryStreamBase<T>) {
+    super(stream.root);
+    this.s = stream;
+  }
+
+  subscribeOne(f: SubscribeOneFn): void {
+    this.s.subscribe((isDone, value, tx) => {
+      if (this.gotValue) {
+        return;
+      }
+      if (isDone) {
+          f(undefined, tx);
+      } else {
+        f(value, tx);
+      }
+      this.gotValue = true;
+    });
+  }
+}
+
+class MapQueryValue<T, S> extends BaseQueryValue<S> {
+  constructor(private v: BaseQueryValue<T>, private mapFn: (x: T) => S) {
+    super(v.root);
+  }
+
+  subscribeOne(f: SubscribeOneFn): void {
+    this.v.subscribeOne((v, tx) => this.mapFn(v));
+  }
+}
+
+
+/**
+ * Exception that should be thrown by client code to abort a transaction.
+ */
+export const AbortTransaction = Symbol("abort_transaction");
 
 /**
  * Get an unresolved promise together with its extracted resolve / reject
  * function.
  */
-export function openPromise<T>() {
+export function openPromise<T>(): any {
   let resolve: ((value?: T | PromiseLike<T>) => void) | null = null;
   let reject: ((reason?: any) => void) | null = null;
   const promise = new Promise<T>((res, rej) => {
@@ -105,41 +243,42 @@ abstract class QueryStreamBase<T> implements QueryStream<T>, PromiseLike<void> {
   abstract subscribe(f: (isDone: boolean,
                          value: any,
                          tx: IDBTransaction) => void): void;
-
-  root: QueryRoot;
-
-  constructor(root: QueryRoot) {
-    this.root = root;
+  constructor(public root: QueryRoot) {
   }
 
-  then<R>(onfulfilled: (value: void) => R | PromiseLike<R>, onrejected: (reason: any) => R | PromiseLike<R>): PromiseLike<R>  {
+  first(): QueryValue<T> {
+    return new FirstQueryValue(this);
+  }
+
+  then<R>(onfulfilled: (value: void) => R | PromiseLike<R>,
+          onrejected: (reason: any) => R | PromiseLike<R>): PromiseLike<R>  {
     return this.root.then(onfulfilled, onrejected);
   }
 
   flatMap<S>(f: (x: T) => S[]): QueryStream<S> {
-    return new QueryStreamFlatMap<T,S>(this, f);
+    return new QueryStreamFlatMap<T, S>(this, f);
   }
 
   map<S>(f: (x: T) => S): QueryStream<S> {
     return new QueryStreamMap(this, f);
   }
 
-  indexJoin<S,I extends IDBValidKey>(index: Index<I,S>,
-                                     keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>> {
+  indexJoin<S, I extends IDBValidKey>(index: Index<I, S>,
+                                      keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>> {
     this.root.addStoreAccess(index.storeName, false);
-    return new QueryStreamIndexJoin(this, index.storeName, index.indexName, keyFn);
+    return new QueryStreamIndexJoin<T, S>(this, index.storeName, index.indexName, keyFn);
   }
 
-  indexJoinLeft<S,I extends IDBValidKey>(index: Index<I,S>,
-                                     keyFn: (obj: T) => I): QueryStream<JoinLeftResult<T, S>> {
+  indexJoinLeft<S, I extends IDBValidKey>(index: Index<I, S>,
+                                          keyFn: (obj: T) => I): QueryStream<JoinLeftResult<T, S>> {
     this.root.addStoreAccess(index.storeName, false);
-    return new QueryStreamIndexJoinLeft(this, index.storeName, index.indexName, keyFn);
+    return new QueryStreamIndexJoinLeft<T, S>(this, index.storeName, index.indexName, keyFn);
   }
 
   keyJoin<S, I extends IDBValidKey>(store: Store<S>,
                                     keyFn: (obj: T) => I): QueryStream<JoinResult<T, S>> {
     this.root.addStoreAccess(store.name, false);
-    return new QueryStreamKeyJoin(this, store.name, keyFn);
+    return new QueryStreamKeyJoin<T, S>(this, store.name, keyFn);
   }
 
   filter(f: (x: any) => boolean): QueryStream<T> {
@@ -147,8 +286,8 @@ abstract class QueryStreamBase<T> implements QueryStream<T>, PromiseLike<void> {
   }
 
   toArray(): Promise<T[]> {
-    let {resolve, promise} = openPromise();
-    let values: T[] = [];
+    const {resolve, promise} = openPromise();
+    const values: T[] = [];
 
     this.subscribe((isDone, value) => {
       if (isDone) {
@@ -164,7 +303,7 @@ abstract class QueryStreamBase<T> implements QueryStream<T>, PromiseLike<void> {
   }
 
   reduce<A>(f: (x: any, acc?: A) => A, init?: A): Promise<any> {
-    let {resolve, promise} = openPromise();
+    const {resolve, promise} = openPromise();
     let acc = init;
 
     this.subscribe((isDone, value) => {
@@ -183,19 +322,11 @@ abstract class QueryStreamBase<T> implements QueryStream<T>, PromiseLike<void> {
 
 type FilterFn = (e: any) => boolean;
 type SubscribeFn = (done: boolean, value: any, tx: IDBTransaction) => void;
-
-interface FlatMapFn<T> {
-  (v: T): T[];
-}
+type SubscribeOneFn = (value: any, tx: IDBTransaction) => void;
 
 class QueryStreamFilter<T> extends QueryStreamBase<T> {
-  s: QueryStreamBase<T>;
-  filterFn: FilterFn;
-
-  constructor(s: QueryStreamBase<T>, filterFn: FilterFn) {
+  constructor(public s: QueryStreamBase<T>, public filterFn: FilterFn) {
     super(s.root);
-    this.s = s;
-    this.filterFn = filterFn;
   }
 
   subscribe(f: SubscribeFn) {
@@ -212,14 +343,9 @@ class QueryStreamFilter<T> extends QueryStreamBase<T> {
 }
 
 
-class QueryStreamFlatMap<T,S> extends QueryStreamBase<S> {
-  s: QueryStreamBase<T>;
-  flatMapFn: (v: T) => S[];
-
-  constructor(s: QueryStreamBase<T>, flatMapFn: (v: T) => S[]) {
+class QueryStreamFlatMap<T, S> extends QueryStreamBase<S> {
+  constructor(public s: QueryStreamBase<T>, public flatMapFn: (v: T) => S[]) {
     super(s.root);
-    this.s = s;
-    this.flatMapFn = flatMapFn;
   }
 
   subscribe(f: SubscribeFn) {
@@ -228,23 +354,18 @@ class QueryStreamFlatMap<T,S> extends QueryStreamBase<S> {
         f(true, undefined, tx);
         return;
       }
-      let values = this.flatMapFn(value);
-      for (let v in values) {
-        f(false, value, tx)
+      const values = this.flatMapFn(value);
+      for (const v in values) {
+        f(false, v, tx);
       }
     });
   }
 }
 
 
-class QueryStreamMap<S,T> extends QueryStreamBase<T> {
-  s: QueryStreamBase<S>;
-  mapFn: (v: S) => T;
-
-  constructor(s: QueryStreamBase<S>, mapFn: (v: S) => T) {
+class QueryStreamMap<S, T> extends QueryStreamBase<T> {
+  constructor(public s: QueryStreamBase<S>, public mapFn: (v: S) => T) {
     super(s.root);
-    this.s = s;
-    this.mapFn = mapFn;
   }
 
   subscribe(f: SubscribeFn) {
@@ -253,7 +374,7 @@ class QueryStreamMap<S,T> extends QueryStreamBase<T> {
         f(true, undefined, tx);
         return;
       }
-      let mappedValue = this.mapFn(value);
+      const mappedValue = this.mapFn(value);
       f(false, mappedValue, tx);
     });
   }
@@ -261,18 +382,9 @@ class QueryStreamMap<S,T> extends QueryStreamBase<T> {
 
 
 class QueryStreamIndexJoin<T, S> extends QueryStreamBase<JoinResult<T, S>> {
-  s: QueryStreamBase<T>;
-  storeName: string;
-  key: any;
-  indexName: string;
-
-  constructor(s: QueryStreamBase<T>, storeName: string, indexName: string,
-              key: any) {
+  constructor(public s: QueryStreamBase<T>, public storeName: string, public indexName: string,
+              public key: any) {
     super(s.root);
-    this.s = s;
-    this.storeName = storeName;
-    this.key = key;
-    this.indexName = indexName;
   }
 
   subscribe(f: SubscribeFn) {
@@ -281,33 +393,24 @@ class QueryStreamIndexJoin<T, S> extends QueryStreamBase<JoinResult<T, S>> {
         f(true, undefined, tx);
         return;
       }
-      let s = tx.objectStore(this.storeName).index(this.indexName);
-      let req = s.openCursor(IDBKeyRange.only(this.key(value)));
+      const s = tx.objectStore(this.storeName).index(this.indexName);
+      const req = s.openCursor(IDBKeyRange.only(this.key(value)));
       req.onsuccess = () => {
-        let cursor = req.result;
+        const cursor = req.result;
         if (cursor) {
           f(false, {left: value, right: cursor.value}, tx);
           cursor.continue();
         }
-      }
+      };
     });
   }
 }
 
 
 class QueryStreamIndexJoinLeft<T, S> extends QueryStreamBase<JoinLeftResult<T, S>> {
-  s: QueryStreamBase<T>;
-  storeName: string;
-  key: any;
-  indexName: string;
-
-  constructor(s: QueryStreamBase<T>, storeName: string, indexName: string,
-              key: any) {
+  constructor(public s: QueryStreamBase<T>, public storeName: string, public indexName: string,
+              public key: any) {
     super(s.root);
-    this.s = s;
-    this.storeName = storeName;
-    this.key = key;
-    this.indexName = indexName;
   }
 
   subscribe(f: SubscribeFn) {
@@ -320,7 +423,7 @@ class QueryStreamIndexJoinLeft<T, S> extends QueryStreamBase<JoinLeftResult<T, S
       const req = s.openCursor(IDBKeyRange.only(this.key(value)));
       let gotMatch = false;
       req.onsuccess = () => {
-        let cursor = req.result;
+        const cursor = req.result;
         if (cursor) {
           gotMatch = true;
           f(false, {left: value, right: cursor.value}, tx);
@@ -330,23 +433,16 @@ class QueryStreamIndexJoinLeft<T, S> extends QueryStreamBase<JoinLeftResult<T, S
             f(false, {left: value}, tx);
           }
         }
-      }
+      };
     });
   }
 }
 
 
 class QueryStreamKeyJoin<T, S> extends QueryStreamBase<JoinResult<T, S>> {
-  s: QueryStreamBase<T>;
-  storeName: string;
-  key: any;
-
-  constructor(s: QueryStreamBase<T>, storeName: string,
-              key: any) {
+  constructor(public s: QueryStreamBase<T>, public storeName: string,
+              public key: any) {
     super(s.root);
-    this.s = s;
-    this.storeName = storeName;
-    this.key = key;
   }
 
   subscribe(f: SubscribeFn) {
@@ -355,17 +451,17 @@ class QueryStreamKeyJoin<T, S> extends QueryStreamBase<JoinResult<T, S>> {
         f(true, undefined, tx);
         return;
       }
-      let s = tx.objectStore(this.storeName);
-      let req = s.openCursor(IDBKeyRange.only(this.key(value)));
+      const s = tx.objectStore(this.storeName);
+      const req = s.openCursor(IDBKeyRange.only(this.key(value)));
       req.onsuccess = () => {
-        let cursor = req.result;
+        const cursor = req.result;
         if (cursor) {
-          f(false, {left:value, right: cursor.value}, tx);
+          f(false, {left: value, right: cursor.value}, tx);
           cursor.continue();
         } else {
           f(true, undefined, tx);
         }
-      }
+      };
     });
   }
 }
@@ -382,7 +478,7 @@ class IterQueryStream<T> extends QueryStreamBase<T> {
     this.storeName = storeName;
     this.subscribers = [];
 
-    let doIt = (tx: IDBTransaction) => {
+    const doIt = (tx: IDBTransaction) => {
       const {indexName = void 0, only = void 0} = this.options;
       let s: any;
       if (indexName !== void 0) {
@@ -391,24 +487,24 @@ class IterQueryStream<T> extends QueryStreamBase<T> {
       } else {
         s = tx.objectStore(this.storeName);
       }
-      let kr: IDBKeyRange | undefined = undefined;
+      let kr: IDBKeyRange | undefined;
       if (only !== undefined) {
         kr = IDBKeyRange.only(this.options.only);
       }
-      let req = s.openCursor(kr);
+      const req = s.openCursor(kr);
       req.onsuccess = () => {
-        let cursor: IDBCursorWithValue = req.result;
+        const cursor: IDBCursorWithValue = req.result;
         if (cursor) {
-          for (let f of this.subscribers) {
+          for (const f of this.subscribers) {
             f(false, cursor.value, tx);
           }
           cursor.continue();
         } else {
-          for (let f of this.subscribers) {
+          for (const f of this.subscribers) {
             f(true, undefined, tx);
           }
         }
-      }
+      };
     };
 
     this.root.addWork(doIt);
@@ -420,9 +516,11 @@ class IterQueryStream<T> extends QueryStreamBase<T> {
 }
 
 
+/**
+ * Root wrapper around an IndexedDB for queries with a fluent interface.
+ */
 export class QueryRoot implements PromiseLike<void> {
-  private work: ((t: IDBTransaction) => void)[] = [];
-  private db: IDBDatabase;
+  private work: Array<((t: IDBTransaction) => void)> = [];
   private stores = new Set();
   private kickoffPromise: Promise<void>;
 
@@ -434,21 +532,37 @@ export class QueryRoot implements PromiseLike<void> {
 
   private finishScheduled: boolean;
 
-  constructor(db: IDBDatabase) {
-    this.db = db;
+  private finished: boolean = false;
+
+  constructor(public db: IDBDatabase) {
   }
 
-  then<R>(onfulfilled: (value: void) => R | PromiseLike<R>, onrejected: (reason: any) => R | PromiseLike<R>): PromiseLike<R> {
+  then<R>(onfulfilled: (value: void) => R | PromiseLike<R>,
+          onrejected: (reason: any) => R | PromiseLike<R>): PromiseLike<R> {
     return this.finish().then(onfulfilled, onrejected);
   }
 
-  iter<T>(store: Store<T>): QueryStream<T> {
-    this.stores.add(store.name);
-    this.scheduleFinish();
-    return new IterQueryStream(this, store.name, {});
+  private checkFinished() {
+    if (this.finished) {
+      throw Error("Can't add work to query after it was started");
+    }
   }
 
+  /**
+   * Get a stream of all objects in the store.
+   */
+  iter<T>(store: Store<T>): QueryStream<T> {
+    this.checkFinished();
+    this.stores.add(store.name);
+    this.scheduleFinish();
+    return new IterQueryStream<T>(this, store.name, {});
+  }
+
+  /**
+   * Count the number of objects in a store.
+   */
   count<T>(store: Store<T>): Promise<number> {
+    this.checkFinished();
     const {resolve, promise} = openPromise();
 
     const doCount = (tx: IDBTransaction) => {
@@ -457,7 +571,7 @@ export class QueryRoot implements PromiseLike<void> {
       req.onsuccess = () => {
         resolve(req.result);
       };
-    }
+    };
 
     this.addWork(doCount, store.name, false);
     return Promise.resolve()
@@ -466,32 +580,37 @@ export class QueryRoot implements PromiseLike<void> {
 
   }
 
+  /**
+   * Delete all objects in a store that match a predicate.
+   */
   deleteIf<T>(store: Store<T>, predicate: (x: T, n: number) => boolean): QueryRoot {
+    this.checkFinished();
     const doDeleteIf = (tx: IDBTransaction) => {
       const s = tx.objectStore(store.name);
       const req = s.openCursor();
       let n = 0;
       req.onsuccess = () => {
-        let cursor: IDBCursorWithValue = req.result;
+        const cursor: IDBCursorWithValue = req.result;
         if (cursor) {
           if (predicate(cursor.value, n++)) {
             cursor.delete();
           }
           cursor.continue();
-        } 
-      }
+        }
+      };
     };
     this.addWork(doDeleteIf, store.name, true);
     return this;
   }
 
-  iterIndex<S extends IDBValidKey,T>(index: Index<S,T>,
-                                     only?: S): QueryStream<T> {
+  iterIndex<S extends IDBValidKey, T>(index: Index<S, T>,
+                                      only?: S): QueryStream<T> {
+    this.checkFinished();
     this.stores.add(index.storeName);
     this.scheduleFinish();
-    return new IterQueryStream(this, index.storeName, {
+    return new IterQueryStream<T>(this, index.storeName, {
+      indexName: index.indexName,
       only,
-      indexName: index.indexName
     });
   }
 
@@ -501,7 +620,8 @@ export class QueryRoot implements PromiseLike<void> {
    * in the store.
    */
   put<T>(store: Store<T>, val: T): QueryRoot {
-    let doPut = (tx: IDBTransaction) => {
+    this.checkFinished();
+    const doPut = (tx: IDBTransaction) => {
       tx.objectStore(store.name).put(val);
     };
     this.scheduleFinish();
@@ -511,12 +631,13 @@ export class QueryRoot implements PromiseLike<void> {
 
 
   putWithResult<T>(store: Store<T>, val: T): Promise<IDBValidKey> {
+    this.checkFinished();
     const {resolve, promise} = openPromise();
-    let doPutWithResult = (tx: IDBTransaction) => {
-      let req = tx.objectStore(store.name).put(val);
+    const doPutWithResult = (tx: IDBTransaction) => {
+      const req = tx.objectStore(store.name).put(val);
       req.onsuccess = () => {
         resolve(req.result);
-      }
+      };
       this.scheduleFinish();
     };
     this.addWork(doPutWithResult, store.name, true);
@@ -526,16 +647,20 @@ export class QueryRoot implements PromiseLike<void> {
   }
 
 
+  /**
+   * Get, modify and store an element inside a transaction.
+   */
   mutate<T>(store: Store<T>, key: any, f: (v: T) => T): QueryRoot {
-    let doPut = (tx: IDBTransaction) => {
-      let reqGet = tx.objectStore(store.name).get(key);
+    this.checkFinished();
+    const doPut = (tx: IDBTransaction) => {
+      const reqGet = tx.objectStore(store.name).get(key);
       reqGet.onsuccess = () => {
-        let r = reqGet.result;
+        const r = reqGet.result;
         let m: T;
         try {
           m = f(r);
         } catch (e) {
-          if (e == AbortTransaction) {
+          if (e === AbortTransaction) {
             tx.abort();
             return;
           }
@@ -543,7 +668,7 @@ export class QueryRoot implements PromiseLike<void> {
         }
 
         tx.objectStore(store.name).put(m);
-      }
+      };
     };
     this.scheduleFinish();
     this.addWork(doPut, store.name, true);
@@ -557,8 +682,9 @@ export class QueryRoot implements PromiseLike<void> {
    * in the object store.
    */
   putAll<T>(store: Store<T>, iterable: T[]): QueryRoot {
+    this.checkFinished();
     const doPutAll = (tx: IDBTransaction) => {
-      for (let obj of iterable) {
+      for (const obj of iterable) {
         tx.objectStore(store.name).put(obj);
       }
     };
@@ -573,6 +699,7 @@ export class QueryRoot implements PromiseLike<void> {
    * in the object store.
    */
   add<T>(store: Store<T>, val: T): QueryRoot {
+    this.checkFinished();
     const doAdd = (tx: IDBTransaction) => {
       tx.objectStore(store.name).add(val);
     };
@@ -585,6 +712,7 @@ export class QueryRoot implements PromiseLike<void> {
    * Get one object from a store by its key.
    */
   get<T>(store: Store<T>, key: any): Promise<T|undefined> {
+    this.checkFinished();
     if (key === void 0) {
       throw Error("key must not be undefined");
     }
@@ -607,13 +735,14 @@ export class QueryRoot implements PromiseLike<void> {
   /**
    * Get one object from a store by its key.
    */
-  getIndexed<I extends IDBValidKey,T>(index: Index<I,T>,
-                                      key: I): Promise<T|undefined> {
+  getIndexed<I extends IDBValidKey, T>(index: Index<I, T>,
+                                       key: I): Promise<T|undefined> {
+    this.checkFinished();
     if (key === void 0) {
       throw Error("key must not be undefined");
     }
 
-    const {resolve, promise} = openPromise();
+    const {resolve, promise} = openPromise<void>();
 
     const doGetIndexed = (tx: IDBTransaction) => {
       const req = tx.objectStore(index.storeName)
@@ -645,7 +774,9 @@ export class QueryRoot implements PromiseLike<void> {
       return this.kickoffPromise;
     }
     this.kickoffPromise = new Promise<void>((resolve, reject) => {
-      if (this.work.length == 0) {
+      // At this point, we can't add any more work
+      this.finished = true;
+      if (this.work.length === 0) {
         resolve();
         return;
       }
@@ -657,7 +788,7 @@ export class QueryRoot implements PromiseLike<void> {
       tx.onabort = () => {
         reject(Error("transaction aborted"));
       };
-      for (let w of this.work) {
+      for (const w of this.work) {
         w(tx);
       }
     });
@@ -668,6 +799,7 @@ export class QueryRoot implements PromiseLike<void> {
    * Delete an object by from the given object store.
    */
   delete(storeName: string, key: any): QueryRoot {
+    this.checkFinished();
     const doDelete = (tx: IDBTransaction) => {
       tx.objectStore(storeName).delete(key);
     };
